from sqlite3 import connect, Connection, Cursor
from . import PYTHON_TYPE_SQLITE, UNUSED, ColumnInfo, camel_cased_to_underlined, class_to_sql

class SqliteTable:
    @classmethod
    def create_table(cls, database:str|Connection, table:str=None, drop:bool=False):
        # 记录默认数据库
        @classmethod
        def sqlite_database(_) -> str|Connection:
            return database
        cls.sqlite_database = sqlite_database
        # 记录数据表的名称
        if table is None:
            table = camel_cased_to_underlined(cls.__name__)
        @classmethod
        def sqlite_table(_) -> str:
            return table
        cls.sqlite_table = sqlite_table
        # 执行数据表的CREATE语句
        is_connection:bool = isinstance(database, Connection)
        connection:Connection  = database if is_connection else connect(database)
        if drop:
            connection.execute(f'DROP TABLE IF EXISTS "{table}";')
        connection.execute(class_to_sql(cls, table))
        connection.commit()
        if not is_connection:
            connection.close()

    @classmethod
    def drop_table(cls, database:str|Connection=None):
        database = cls.sqlite_database() if database is None else database
        # 执行数据表的DROP语句
        is_connection:bool = isinstance(database, Connection)
        connection:Connection  = database if is_connection else connect(database)
        connection.execute(f'DROP TABLE IF EXISTS "{cls.sqlite_table()}";')
        connection.commit()
        if not is_connection:
            connection.close()

    def sqlite_insert(self, database:str|Connection=None) -> int:
        cls = self.__class__
        database = cls.sqlite_database() if database is None else database
        lastrowid:int = None
        # 列名容器
        cols:list[str] = []
        vals:list[str] = []
        # 获取插入的键和值
        for n in cls.__annotations__.keys():
            v = getattr(self, n, UNUSED)
            if v!=UNUSED:
                cols.append(f'"{n}"')
                vals.append(f':{n}')
        # 执行数据表的INSERT语句
        is_connection:bool = isinstance(database, Connection)
        connection:Connection  = database if is_connection else connect(database)
        lastrowid = connection.execute(f'INSERT INTO "{cls.sqlite_table()}" ({",".join(cols)}) VALUES ({",".join(vals)});', self.__dict__).lastrowid
        connection.commit()
        if not is_connection:
            connection.close()
        return lastrowid

    def sqlite_select(self, database:str|Connection=None) -> 'SqliteTable':
        cls = self.__class__
        database = cls.sqlite_database() if database is None else database
        cols:list[str] = []
        # 条件语句容器
        wheres:list[str] = []
        # 获取条件的键和值
        for n in cls.__annotations__.keys():
            v = getattr(self, n, UNUSED)
            if v != UNUSED:
                wheres.append(f'"{n}"=:{n}')
            cols.append(n)
        # 执行数据表的SELECT语句
        is_connection:bool = isinstance(database, Connection)
        connection:Connection  = database if is_connection else connect(database)
        cursor:Cursor = connection.execute(f'''SELECT {",".join((f'"{n}"' for n in cols))} FROM "{cls.sqlite_table()}"{" WHERE " if wheres else ""}{" AND ".join(wheres)};''', self.__dict__)
        for row in cursor:
            obj = cls()
            for n, v in zip(cols, row):
                setattr(obj,n,v)
            yield obj
        if not is_connection:
            connection.close()

    @classmethod
    def sqlite_database(_) -> str|Connection:
        """ 获取创建数据表时使用的数据库 """
        return None

    @classmethod
    def sqlite_table(_) -> str:
        """ 获取创建数据表时使用的数据表名称 """
        return None




from . import PYTHON_TYPE_SQLITE, ColumnInfo
from re import sub

def camel_cased_to_underlined(name:str) -> str:
    """ 驼峰转下划线 """
    return sub(r'([a-z])([A-Z])', r'\1_\2', name).lower()

def class_to_sql(cls:type, table:str) -> str:
    """ 从类cls中获取数据表列信息并生产创建该数据表table的sql语句 """
    # 列sql容器
    cols:list[str] = []
    # PRIMARY KEY容器
    pks:list[str] = []
    # 遍历类注解，获取列信息
    for n, t in cls.__annotations__.items():
        # 提取类变量名作为列名、类变量注解作为列类型
        v = getattr(cls, n, ColumnInfo(name=n, type=PYTHON_TYPE_SQLITE[t]))
        # 当类变量的值不是ColumnInfo时将该值作为列的默认取值
        if not isinstance(v, ColumnInfo):
            v = ColumnInfo(name=n, type=PYTHON_TYPE_SQLITE[t], dflt_value='NULL' if v is None else str(v))
        # 转换类变量值为UNUSED
        setattr(cls, n, v)
        # 组装列sql
        cols.append(f'"{v.name}" {v.type}{" NOT NULL" if v.notnull else ""}{f" DEFAULT {v.dflt_value}" if v.dflt_value is not None else ""}')
        # 记录PRIMARY KEY
        if v.pk:
            pks.append(f'"{v.name}"')
    # 组装sql语句
    return f'''CREATE TABLE IF NOT EXISTS "{table}"({",".join(cols)}{f',CONSTRAINT "pk_{table}" PRIMARY KEY ({",".join(pks)})' if pks else ""});'''



from ._Constants import PYTHON_TYPE_SQLITE, UNUSED
from ._ColumnInfo import ColumnInfo
from ._SqlTools import camel_cased_to_underlined, class_to_sql
from ._SqliteTable import SqliteTable

name = 'sqlitetable'
__version__ = '0.0.2'
VERSION = __version__
__all__ = [
    'UNUSED',
    'ColumnInfo',
    'SqliteTable'
]



from . import PYTHON_TYPE_SQLITE, UNUSED

class ColumnInfo:
    """ 数据表的列信息，对应sql语句 f'pragma table_info("{table}")' 的返回结果 """
    def __init__(self,cid:int = None,name:str = None,type:str = None,notnull:int = 0,dflt_value:None|str = None,pk:int = 0) -> None:
        self.cid:int = cid
        self.name:str = name
        self.type:str = type
        self.notnull:int = notnull
        self.dflt_value:None|str = dflt_value
        self.pk:int = pk

    def __set_name__(self, cls:type, name:str):
        # 提取类变量名作为列名
        if self.name is None:
            self.name = name
        # 提取类变量注解作为列类型
        if self.type is None:
            self.type = PYTHON_TYPE_SQLITE[cls.__annotations__[name]]
    
    def __get__(self, obj:object, cls:type=None) -> any:
        if obj is None:
            return self
        if self.name in obj.__dict__:
            return obj.__dict__[self.name]
        return UNUSED

    def __set__(self, obj:object, value):
        obj.__dict__[self.name] = value


class Sign:
    """ 标记，用于创建不冲突的常量 """
    def __init__(self, name:str) -> None:
        self.name = name

# 数据表列没有被使用
UNUSED = Sign('UnUsed')

# python类型对应的sqlite类型
PYTHON_TYPE_SQLITE = {
    None: "NULL",
    int: "INTEGER",
    float: "REAL",
    str: "TEXT",
    bytes: "BLOB"
}


from setuptools import setup, find_packages
import sqlitetable

setup(
    name='sqlitetable',
    version=sqlitetable.__version__,
    description=(
        'Use sqlite just like classes!'
    ),
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    long_description_content_type="text/markdown",
    author='Poiuzy',
    author_email='poiuzy-ice@qq.com',
    maintainer='Poiuzy',
    maintainer_email='poiuzy-ice@qq.com',
    license='MIT License',
    packages=find_packages(),
    platforms=["all"]
)

